helps = dict(
    run = '''run shell cmd on target
    ''',
    module = '''show module's help
    ''',
    modules = '''list modules
        arg = search pattern
    ''',
    deploy = '''deploy module
    ''',
    remote = '''deploy module remotly
    ''',
    update = '''update packages
    ''',
    upgrade = '''upgrade packages
        May hang, if user input is required.
    ''',
    installed = '''list installed packages
    ''',
    available = '''list avaiable packages
    ''',
    bigs = '''list the bigest installed packages
    ''',
    upgradable = '''list upgradble packages
    ''',
    install = '''install package
    ''',
    remove = '''remove package
    ''',
    cleanup = '''cleanup packages
    ''',
    status = '''query service status
    ''',
    start = '''start service
    ''',
    stop = '''stop service
    ''',
    restart = '''restart service
    ''',
    enable = '''enable service
    ''',
    disable = '''disable service
    ''',
    download = '''download file
    ''',
    logs = '''show log file
        arg = level (default to "info")
    ''',
    outrun = '''show commands history
    ''',
    outlog = '''show commands log
    ''',
)



from setux.main import banner


doc = f'''{banner}

setux [Module | Command] [*args | **kwargs] [Target]

Deploy Module or Execute Command on Target


Module or Command:
    - Deploy Module ( see the "modules" command )
    - Execute Command ( see the "help" command )
    - if not specified : enter REPL on Target

Target:
    - May be passed on command line
    - Set in environement as "setux_target"
    - defaults to "Local"
'''


def usage(*args):
    print(doc)



from os import environ
from os.path import expanduser
from sys import argv

from setux.targets import Local, SSH
from setux.logger import debug
from setux.repl import commands
from setux.repl.repl import repl, help
from .usage import usage


def get_target(name=None, dest=None):
    name = name or environ.get('setux_target')
    dest = dest or environ.get('setux_outdir', expanduser('~/setux'))
    if name:
        target = SSH(name=name, host=name, outdir=dest)
        if not target.cnx:
            target = None
    else:
        target = Local(outdir=dest)
    return target


def main():

    if len(argv)==1:
        target = get_target()
        debug(f'repl {target}')
        repl(target)


    else:
        del argv[0]
        try: target = get_target(argv[-1])
        except: target = None
        if target:
            del argv[-1]
        else:
            target = get_target()

        if argv:
            name, *args = argv
        else:
            repl(target)
            return

        if name in target.modules.items:
            try:
                k = dict(i.split('=') for i in args)
                try:
                    target.deploy(name, **k)
                except KeyError as x:
                    key = x.args[0]
                    print(f'\n ! missing argument : {key}  !\n')
                    commands['module'](target, name)
            except ValueError:
                m = "module's arguments must be keyword arguments\n"
                print(f'\n ! invalid argument : {" ".join(args)} !\n ! {m}')
                commands['module'](target, name)
            return

        if name=='help':
            cmd = help
            help(args[0] if args else None)
        else:
            cmd = commands.get(name)
            if cmd:
                cmd(target, args[0] if args else None)
            else:
                print(f'\n ! {name} !\n')
                return usage()





def run(target, arg):
    target(arg)

def module(target, arg):
    mod = target.modules.items.get(arg)
    if mod:
        title = f'module {arg}'
        print(title)
        print('-'*len(title))
        print(mod.help())
    else:
        print(f' ! unkown module ! {arg}')

def modules(target, arg):
    modules = target.modules.items
    print('modules')
    print('-------')
    width = len(max(modules.keys(), key=len))+4
    for name, mod in sorted(modules.items()):
        hlp = mod.help()
        first = hlp.split('\n')[0]
        if (
            not arg
            or arg in name
            or arg in first.lower()
        ): print(f'{name:>{width}} {first}')

def deploy(target, arg):
    target.deploy(arg)

def remote(target, arg):
    target.remote(arg)

def update(target, arg):
    target.Package.update()

def upgrade(target, arg):
    target.Package.upgrade()

def installed(target, arg):
    pkg = list(target.Package.installed(arg))
    print('\n'.join(f'{n} {v}' for n, v in pkg))

def available(target, arg):
    pkg = list(target.Package.available(arg))
    print('\n'.join(f'{n} {v}' for n, v in pkg))

def bigs(target, arg):
    pkg = list(target.Package.bigs())
    print('\n'.join(pkg))

def upgradable(target, arg):
    pkg = list(target.Package.upgradable())
    print('\n'.join(n for n, v in pkg))

def install(target, arg):
    target.Package.install(arg)

def remove(target, arg):
    target.Package.remove(arg)

def cleanup(target, arg):
    target.Package.cleanup()

def status(target, arg):
    target.Service.status(arg)

def start(target, arg):
    target.Service.start(arg)

def stop(target, arg):
    target.Service.stop(arg)

def restart(target, arg):
    target.Service.restart(arg)

def enable(target, arg):
    target.Service.enable(arg)

def disable(target, arg):
    target.Service.disable(arg)

def download(target, arg):
    url, dst = arg.split(' ') if ' ' in arg else arg, None
    target.deploy('download', url=url, dst=dst)

def outrun(target, arg):
    log = target.outrun
    if log:
        print(open(log).read())
    else:
        print('target outrun not defined')

def outlog(target, arg):
    log = target.outlog
    if log:
        print(open(log).read())
    else:
        print('target outlog not defined')



import readline
readline.parse_and_bind('tab: complete')
readline.parse_and_bind('set editing-mode vi')

from cmd import Cmd

from setux.main import banner
from setux.logger import logger
from .helps import helps
from . import commands


def help(cmd=None):
    def title(txt):
        print(txt)
        print('-'*len(txt))

    if cmd:
        hlp = helps.get(cmd)
        if hlp:
            title(cmd)
            print(hlp)
        else:
            print(f'unkown command "{cmd}"')
    else:
        title('commands')
        width = len(max(helps.keys(), key=len))+4
        for cmd, hlp in sorted(helps.items()):
            first = hlp.split('\n')[0]
            print(f'{cmd:>{width}} {first}')


def safe(cmd, target):
    def wrapper(arg):
        try:
            cmd(target, arg)
        except Exception as x:
            print(type(x).__name__, x)
    return wrapper


def prompt(target):
    user = 'sudo' if target.sudo else target.login.name
    return f'{user}@{target.name} > '


class Repl(Cmd):
    def __init__(self, target):
        self.prompt = prompt(target)
        for name, cmd in commands.items():
            setattr(self, f'do_{name}', safe(cmd, target))
        self.modules = target.modules.items.keys()
        super().__init__()

    def do_help(self, cmd):
        help(cmd)

    def do_logs(self, level=None):
        log = logger.logs(level or 'info')
        print(open(log).read())

    def preloop(self):
        print(banner)
        self.onecmd('infos')
        print()

    def default(self, line):
        module = line.split()[0]
        if module in self.modules:
            self.onecmd(f'deploy {line}')
        else:
            self.onecmd(f'run {line}')

    def do_EOF(self, arg):
        return True


def repl(target):
    Repl(target).cmdloop()
